use color_eyre::Result;
use crossterm::event::{KeyCode, KeyEvent};
use ratatui::layout::Margin;
use ratatui::widgets::ListItem;
use ratatui::{
    layout::{Alignment, Rect}, 
    widgets::{
        Block, List, ListState, Scrollbar,
        ScrollbarOrientation, ScrollbarState,
    }, 
    Frame
};
use tokio::sync::mpsc::UnboundedSender;

use super::{PartialComponent, Component, CurrentTab};
use crate::{
    action::Action, 
    config::Config, 
    palette::StylePalette, 
};

pub struct ListenAddresses {
    is_active: bool,
    action_tx: Option<UnboundedSender<Action>>,
    palette: StylePalette,
    scroll_state: ScrollbarState,
    list_state: ListState,
    listen_addresses: Vec<String>,
    local_identity: String,
}

impl Default for ListenAddresses {
    fn default() -> Self {
        Self::new()
    }
}

impl ListenAddresses {
    pub fn new() -> Self {
        Self {
            is_active: false,
            action_tx: None,
            scroll_state: ScrollbarState::new(0),
            list_state: ListState::default(),
            palette: StylePalette::default(),
            listen_addresses: Vec::new(),
            local_identity: String::new(),
        }
    }

    fn first_row(&mut self) {
        if self.listen_addresses.len() > 0 {
            self.list_state.select(Some(0));
            self.scroll_state = self.scroll_state.position(0);
        }
    }

    fn next_row(&mut self) {
        let i = match self.list_state.selected() {
            Some(i) => {
                if i >= self.listen_addresses.len() - 1 {
                    i
                } else {
                    i + 1
                }
            },
            None => 0,
        };
        self.list_state.select(Some(i));
        self.scroll_state = self.scroll_state.position(i);
    }

    fn last_row(&mut self) {
        if self.listen_addresses.len() > 0 {
            let last = self.listen_addresses.len() - 1;
            self.list_state.select(Some(last));
            self.scroll_state = self.scroll_state.position(last);
        }
    }

    fn previous_row(&mut self) {
        let i = match self.list_state.selected() {
            Some(i) => {
                if i == 0 {
                    0
                } else {
                    i - 1
                }
            },
            None => 0
        };
        self.list_state.select(Some(i));
        self.scroll_state = self.scroll_state.position(i);
    }
}

impl PartialComponent for ListenAddresses {
    fn set_active(&mut self, current_tab: CurrentTab) {
        match current_tab {
            CurrentTab::ListenAddresses => self.is_active = true,
            _ => {
                self.is_active = false;
                self.list_state.select(None);
                self.scroll_state = self.scroll_state.position(0);
            }
        }
    }
}

impl Component for ListenAddresses {
    fn register_action_handler(&mut self, tx: UnboundedSender<Action>) -> Result<()> {
        self.action_tx = Some(tx);
        Ok(())
    }

    fn register_config_handler(&mut self, config: Config) -> Result<()> {
        if let Some(style) = config.styles.get(&crate::app::Mode::Wallet) {
            self.palette.with_normal_style(style.get("normal_style").copied());
            self.palette.with_hover_style(style.get("hover_style").copied());
            self.palette.with_normal_border_style(style.get("normal_border_style").copied());
            self.palette.with_hover_border_style(style.get("hover_border_style").copied());
            self.palette.with_normal_title_style(style.get("normal_title_style").copied());
            self.palette.with_hover_title_style(style.get("hover_title_style").copied());
            self.palette.with_highlight_style(style.get("highlight_style").copied());
            self.palette.with_scrollbar_style(style.get("scrollbar_style").copied());
        }
        Ok(())
    }

    fn update(&mut self, action: Action) -> Result<Option<Action>> {
        match action {
            Action::SetListenAddresses(addresses) => self.listen_addresses = addresses,
            Action::SetLocalIdentity(identity) => self.local_identity = identity,
            _ => {}
        };
        Ok(None)
    }

    fn handle_key_event(&mut self, key: KeyEvent) -> Result<Option<Action>> {
        if self.is_active {
            match key.code {
                KeyCode::Up | KeyCode::Char('k') => self.previous_row(), 
                KeyCode::Down | KeyCode::Char('j') => self.next_row(), 
                KeyCode::Char('g') => self.first_row(),
                KeyCode::Char('G') => self.last_row(),
                _ => {},
            };
        }
        Ok(None)
    }

    fn draw(&mut self, frame: &mut Frame, area: Rect) -> Result<()> {
        let [_, place] = super::validator_session_and_listen_layout(area);
        let (border_style, border_type) = self.palette.create_border_style(self.is_active);
        let list = List::new(
            self.listen_addresses
                .iter()
                .map(|addr| ListItem::new(addr.clone()))
            )
            .highlight_style(self.palette.create_highlight_style())
            .block(Block::bordered()
                .border_style(border_style)
                .border_type(border_type)
                .title_alignment(Alignment::Right)
                .title_style(self.palette.create_title_style(false))
                .title(self.local_identity.clone()));

        let scrollbar = Scrollbar::default()
            .orientation(ScrollbarOrientation::VerticalRight)
            .begin_symbol(None)
            .end_symbol(None)
            .style(self.palette.create_scrollbar_style());

        frame.render_stateful_widget(list, place, &mut self.list_state);
        frame.render_stateful_widget(
            scrollbar,
            place.inner(Margin { vertical: 1, horizontal: 1 }),
            &mut self.scroll_state,
        );

        Ok(())
    }
}
