#![allow(dead_code)]

use std::cmp::min;

use derive_builder::Builder;
use font8x8::UnicodeFonts;
use ratatui::{prelude::*, text::StyledGrapheme, widgets::Widget};

#[derive(Debug, Builder, Clone, PartialEq, Eq, Hash)]
#[builder(build_fn(skip))]
pub struct BigText<'a> {
    #[builder(default, setter(into))]
    pub lines: Vec<Line<'a>>,

    #[builder(default, setter(into))]
    pub style: Style,

    #[builder(default)]
    pub pixel_size: PixelSize,

    #[builder(default)]
    pub alignment: Alignment,
}

impl BigText<'static> {
    pub fn builder() -> BigTextBuilder<'static> {
        BigTextBuilder::default()
    }
}

#[derive(Debug, Clone, Copy, Eq, PartialEq, Hash, Default)]
pub enum PixelSize {
    #[default]
    Full,
    HalfHeight,
    HalfWidth,
    Quadrant,
    ThirdHeight,
    Sextant,
}

impl PixelSize {
    pub(crate) fn pixels_per_cell(self) -> (u16, u16) {
        match self {
            PixelSize::Full => (1, 1),
            PixelSize::HalfHeight => (1, 2),
            PixelSize::HalfWidth => (2, 1),
            PixelSize::Quadrant => (2, 2),
            PixelSize::ThirdHeight => (1, 3),
            PixelSize::Sextant => (2, 3),
        }
    }

    pub(crate) fn symbol_for_position(self, glyph: &[u8; 8], row: usize, col: i32) -> char {
        match self {
            PixelSize::Full => match glyph[row] & (1 << col) {
                0 => ' ',
                _ => '█',
            },
            PixelSize::HalfHeight => {
                let top = glyph[row] & (1 << col);
                let bottom = glyph[row + 1] & (1 << col);
                get_symbol_half_height(top, bottom)
            }
            PixelSize::HalfWidth => {
                let left = glyph[row] & (1 << col);
                let right = glyph[row] & (1 << (col + 1));
                get_symbol_half_width(left, right)
            }
            PixelSize::Quadrant => {
                let top_left = glyph[row] & (1 << col);
                let top_right = glyph[row] & (1 << (col + 1));
                let bottom_left = glyph[row + 1] & (1 << col);
                let bottom_right = glyph[row + 1] & (1 << (col + 1));
                get_symbol_quadrant_size(top_left, top_right, bottom_left, bottom_right)
            }
            PixelSize::ThirdHeight => {
                let top = glyph[row] & (1 << col);
                let is_middle_available = (row + 1) < glyph.len();
                let middle = if is_middle_available {
                    glyph[row + 1] & (1 << col)
                } else {
                    0
                };
                let is_bottom_available = (row + 2) < glyph.len();
                let bottom = if is_bottom_available {
                    glyph[row + 2] & (1 << col)
                } else {
                    0
                };
                get_symbol_third_height(top, middle, bottom)
            }
            PixelSize::Sextant => {
                let top_left = glyph[row] & (1 << col);
                let top_right = glyph[row] & (1 << (col + 1));
                let is_middle_available = (row + 1) < glyph.len();
                let (middle_left, middle_right) = if is_middle_available {
                    (
                        glyph[row + 1] & (1 << col),
                        glyph[row + 1] & (1 << (col + 1)),
                    )
                } else {
                    (0, 0)
                };
                let is_bottom_available = (row + 2) < glyph.len();
                let (bottom_left, bottom_right) = if is_bottom_available {
                    (
                        glyph[row + 2] & (1 << col),
                        glyph[row + 2] & (1 << (col + 1)),
                    )
                } else {
                    (0, 0)
                };
                get_symbol_sextant_size(
                    top_left,
                    top_right,
                    middle_left,
                    middle_right,
                    bottom_left,
                    bottom_right,
                )
            }
        }
    }
}

fn get_symbol_half_height(top: u8, bottom: u8) -> char {
    match top {
        0 => match bottom {
            0 => ' ',
            _ => '▄',
        },
        _ => match bottom {
            0 => '▀',
            _ => '█',
        },
    }
}

fn get_symbol_half_width(left: u8, right: u8) -> char {
    match left {
        0 => match right {
            0 => ' ',
            _ => '▐',
        },
        _ => match right {
            0 => '▌',
            _ => '█',
        },
    }
}

fn get_symbol_quadrant_size(
    top_left: u8,
    top_right: u8,
    bottom_left: u8,
    bottom_right: u8,
) -> char {
    let top_left = if top_left > 0 { 1 } else { 0 };
    let top_right = if top_right > 0 { 1 } else { 0 };
    let bottom_left = if bottom_left > 0 { 1 } else { 0 };
    let bottom_right = if bottom_right > 0 { 1 } else { 0 };

    const QUADRANT_SYMBOLS: [char; 16] = [
        ' ', '▘', '▝', '▀', '▖', '▌', '▞', '▛', '▗', '▚', '▐', '▜', '▄', '▙', '▟', '█',
    ];
    let character_index = top_left + (top_right << 1) + (bottom_left << 2) + (bottom_right << 3);

    QUADRANT_SYMBOLS[character_index]
}

fn get_symbol_third_height(top: u8, middle: u8, bottom: u8) -> char {
    get_symbol_sextant_size(top, top, middle, middle, bottom, bottom)
}

fn get_symbol_sextant_size(
    top_left: u8,
    top_right: u8,
    middle_left: u8,
    middle_right: u8,
    bottom_left: u8,
    bottom_right: u8,
) -> char {
    let top_left = if top_left > 0 { 1 } else { 0 };
    let top_right = if top_right > 0 { 1 } else { 0 };
    let middle_left = if middle_left > 0 { 1 } else { 0 };
    let middle_right = if middle_right > 0 { 1 } else { 0 };
    let bottom_left = if bottom_left > 0 { 1 } else { 0 };
    let bottom_right = if bottom_right > 0 { 1 } else { 0 };

    const SEXANT_SYMBOLS: [char; 64] = [
        ' ', '🬀', '🬁', '🬂', '🬃', '🬄', '🬅', '🬆', '🬇', '🬈', '🬉', '🬊', '🬋', '🬌', '🬍', '🬎', '🬏', '🬐',
        '🬑', '🬒', '🬓', '▌', '🬔', '🬕', '🬖', '🬗', '🬘', '🬙', '🬚', '🬛', '🬜', '🬝', '🬞', '🬟', '🬠', '🬡',
        '🬢', '🬣', '🬤', '🬥', '🬦', '🬧', '▐', '🬨', '🬩', '🬪', '🬫', '🬬', '🬭', '🬮', '🬯', '🬰', '🬱', '🬲',
        '🬳', '🬴', '🬵', '🬶', '🬷', '🬸', '🬹', '🬺', '🬻', '█',
    ];
    let character_index = top_left
        + (top_right << 1)
        + (middle_left << 2)
        + (middle_right << 3)
        + (bottom_left << 4)
        + (bottom_right << 5);

    SEXANT_SYMBOLS[character_index]
}

impl BigTextBuilder<'_> {
    pub fn left_aligned(&mut self) -> &mut Self {
        self.alignment(Alignment::Left)
    }
    pub fn centered(&mut self) -> &mut Self {
        self.alignment(Alignment::Center)
    }
    pub fn right_aligned(&mut self) -> &mut Self {
        self.alignment(Alignment::Right)
    }
}

impl<'a> BigTextBuilder<'a> {
    pub fn build(&self) -> BigText<'a> {
        BigText {
            lines: match &self.lines {
                Some(lines) => lines.clone(),
                None => Vec::new(),
            },
            style: match &self.style {
                Some(style) => *style,
                None => Style::default(),
            },
            pixel_size: self.pixel_size.unwrap_or_default(),
            alignment: self.alignment.unwrap_or_default(),
        }
    }
}

impl Widget for BigText<'_> {
    fn render(self, area: Rect, buf: &mut Buffer) {
        let layout = layout(area, &self.pixel_size, self.alignment, &self.lines);
        for (line, line_layout) in self.lines.iter().zip(layout) {
            for (g, cell) in line.styled_graphemes(self.style).zip(line_layout) {
                render_symbol(g, cell, buf, &self.pixel_size);
            }
        }
    }
}

fn layout<'a>(
    area: Rect,
    pixel_size: &PixelSize,
    alignment: Alignment,
    lines: &'a [Line<'a>],
) -> impl IntoIterator<Item = impl IntoIterator<Item = Rect>> + 'a {
    let (step_x, step_y) = pixel_size.pixels_per_cell();
    let width = 8u16.div_ceil(step_x);
    let height = 8u16.div_ceil(step_y);

    (area.top()..area.bottom())
        .step_by(height as usize)
        .zip(lines.iter())
        .map(move |(y, line)| {
            let offset = get_alignment_offset(area.width, width, alignment, line);
            (area.left() + offset..area.right())
                .step_by(width as usize)
                .map(move |x| {
                    let width = min(area.right() - x, width);
                    let height = min(area.bottom() - y, height);
                    Rect::new(x, y, width, height)
                })
        })
}

fn get_alignment_offset<'a>(
    area_width: u16,
    letter_width: u16,
    alignment: Alignment,
    line: &'a Line<'a>,
) -> u16 {
    let big_line_width = line.width() as u16 * letter_width;
    match alignment {
        Alignment::Left => 0,
        Alignment::Center => (area_width / 2).saturating_sub(big_line_width / 2),
        Alignment::Right => area_width.saturating_sub(big_line_width),
    }
}

fn render_symbol(grapheme: StyledGrapheme, area: Rect, buf: &mut Buffer, pixel_size: &PixelSize) {
    buf.set_style(area, grapheme.style);
    let c = grapheme.symbol.chars().next().unwrap();
    if let Some(glyph) = font8x8::BASIC_FONTS.get(c) {
        render_glyph(glyph, area, buf, pixel_size);
    }
}

fn render_glyph(glyph: [u8; 8], area: Rect, buf: &mut Buffer, pixel_size: &PixelSize) {
    let (step_x, step_y) = pixel_size.pixels_per_cell();

    let glyph_vertical_index = (0..glyph.len()).step_by(step_y as usize);
    let glyph_horizontal_bit_index = (0..8).step_by(step_x as usize);

    for (y, row) in glyph_vertical_index.zip(area.rows()) {
        for (x, col) in glyph_horizontal_bit_index.clone().zip(row.columns()) {
            buf[col].set_char(pixel_size.symbol_for_position(&glyph, y, x));
        }
    }
}
